;
; Concentration Room game logic
; Copyright (C) 2010 Damian Yerrick
;
; This program is free software: you can redistribute it and/or modify
; it under the terms of the GNU General Public License as published by
; the Free Software Foundation, either version 3 of the License, or
; (at your option) any later version.
;
; This program is distributed in the hope that it will be useful,
; but WITHOUT ANY WARRANTY; without even the implied warranty of
; MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
; GNU General Public License for more details.
;
; You should have received a copy of the GNU General Public License
; along with this program.  If not, see <http://www.gnu.org/licenses/>.
;
.include "src/nes.h"
.include "src/ram.h"

USE_SELECT_FOR_SLOWDOWN = 1
USE_B_FOR_RANDOM = 1

.segment "ZEROPAGE"
curState: .res 1
stateTimer: .res 1
cardToDraw: .res 1  ; card to redraw next frame
isCombo: .res 1  ; nonzero if the last was a combo
bgcolor: .res 1
gameOverClearTransitionY: .res 1
difficulty: .res 1
numPlayers: .res 1
lastPlayerIsAI: .res 1
scoreMethod: .res 1  ; 0: decrement on miss; 1: increment on match
score: .res 2  ; one for each player

; The game supports passing 1 controller or using 2 controllers.
activePad: .res 1  ; shows which controller the player is using
das_timer: .res 2  ; one for each pad
das_keys: .res 2  ; one for each pad

.segment "CODE"
.proc play_memory
  lda #0
  sta curTurn
  sta gameOverClearTransitionY
  sta curAIState
  sta isCombo
  sta activePad
  jsr shuffleCards
  lda #$18
  sta bgcolor
  jsr loadPlayScreen

  lda #4
  sta cursor_x
  lda #3
  sta cursor_y
  
  lda #$FF
  sta selectedCards
  sta selectedCards+1
  
  ; Draw all the face-down cards to VRAM
  ldx #FIELD_HT*FIELD_WID-1
  stx cardToDraw
surroundsLoop:
  ldx cardToDraw
  lda #0
  sta rememberState,x  ; clear CPU-remembered state of this card
  txa
  jsr buildCardTiles
  jsr blitCard
  dec cardToDraw
  bpl surroundsLoop
  lda #PlayState::STILL
  sta curState
  
cardsLoop:
  jsr read_pads
  ldx activePad
  jsr autorepeat

  lda curState
  cmp #PlayState::PASS_CONTROLLER
  beq notRight
  
  ldx activePad
  lda new_keys,x
  and #KEY_UP
  beq notUp
  lda cursor_y
  beq notUp
  dec cursor_y
notUp:
  lda new_keys,x
  and #KEY_LEFT
  beq notLeft
  lda cursor_x
  beq notLeft
  dec cursor_x
notLeft:
  lda new_keys,x
  and #KEY_DOWN
  beq notDown
  lda cursor_y
  cmp #FIELD_HT - 1
  bcs notDown
  inc cursor_y
notDown:
  lda new_keys,x
  and #KEY_RIGHT
  beq notRight
  lda cursor_x
  cmp #FIELD_WID - 1
  bcs notRight
  inc cursor_x
notRight:
  jsr stateDispatch

  ldx cardToDraw
  bmi :+
  jsr buildCardTiles
:
  jsr drawCardSprites
  jsr update_sound

  ; all done; now wait for vblank and blit the damn things
.if ::USE_SELECT_FOR_SLOWDOWN
  ldx #1
  lda #KEY_SELECT
  and cur_keys
  beq mainSlowdown
  ldx #15
.endif
mainSlowdown:
  lda nmis
:
  cmp nmis
  beq :-
.if ::USE_SELECT_FOR_SLOWDOWN
  dex
  bne mainSlowdown
.endif
  lda #0
  sta PPUMASK
  lda PPUSTATUS
  jsr blitCardSprites
  ldx cardToDraw
  bmi :+
  jsr blitCard
:
  jsr blitScoreUpdate
  jsr gameOverClearRow
  lda #VBLANK_NMI|BG_1000|OBJ_1000
  sta PPUCTRL
  lda #0
  sta PPUSCROLL
  lda #12
  sta PPUSCROLL
  lda #%00011110
  sta PPUMASK
  lda #$FF
  sta cardToDraw
  lda curState
  cmp #PlayState::DONE
  beq exit
  jmp cardsLoop
exit:
  rts
.endproc

.proc stateDispatch
  lda curState
  asl a
  tax
  lda stateHandlers+1,x
  pha
  lda stateHandlers,x
  pha
  rts
  
stateHandlers:
  .addr handleStateStill-1, handleStateFlipping-1
  .addr handleStateUnflipping-1, handleStateCollecting-1
  .addr handleStatePassController-1, handleStateCleared-1
  .addr handleStateGameOver-1
.endproc

.proc handleStateCleared
  ldx curTurn
  lda score,x
  tay
  jsr buildScoreUpdate
  dec stateTimer
  beq done
  rts
done:
  lda #PlayState::DONE
  sta curState
  sec
  sbc #$04
.endproc

.proc handleStateStill
.if ::USE_B_FOR_RANDOM
  ldx activePad
  lda new_keys,x
  and #KEY_B
  beq notB
  lda #1
  jsr start_sound
  jsr randomCard
  cpy #72
  bcc :+
  ldy #71
:
  tya
  and #$07
  sta cursor_y
  tya
  lsr a
  lsr a
  lsr a
  and #$0F
  sta cursor_x
notB:
.endif

  lda #0
  sta card0FlipFrame
  lda lastPlayerIsAI
  beq notAI
  lda numPlayers
  clc
  sbc curTurn
  bne notAI
  jsr doAI
notAI:

  ldx activePad
  lda new_keys,x
  and #KEY_A
  beq notA

  ; make sure the card is flipped over first
  lda cursor_x
  asl a
  asl a
  asl a
  ora cursor_y
  tay
  cpy selectedCards+1  ; if same as other card, skip
  beq notA
  cpy selectedCards  ; if same as this card, skip
  beq notA
  lda boardState,y  ; if no card is there, skip
  bpl notA

  sty selectedCards
  sty cardToDraw
  lda boardState,y
  and #$3F
  ora #$40
  sta boardState,y
  lda #PlayState::FLIPPING
  sta curState
  lda #0
  jsr start_sound
  lda #$08
  sta card0FlipFrame
  lda #5
  sta stateTimer
  bne notA
notA:
  rts
.endproc

.proc handleStateFlipping
  lda stateTimer
  beq readyToFinish
notFirstFrame:  
  dec stateTimer
  lda stateTimer  ; $05-$00
  lsr a           ; $02-$00
  eor #$03        ; $01-$03
  clc
  adc #$07        ; $08-$0A
  sta card0FlipFrame
  rts
readyToFinish:
  ldy selectedCards
  bmi nothingSelected
  sty cardToDraw
  lda boardState,y
  ora #$C0
  sta boardState,y
  sta rememberState,y
  
  ; if the other card isn't flipped, swap these
  ldx selectedCards+1
  bpl hasCard2
  sty selectedCards+1
  ldy #$FF
  sty selectedCards
  lda #PlayState::STILL
  sta curState
nothingSelected:
  lda #0
  sta card0FlipFrame
  rts
hasCard2:

  ; After the player turns over the second card, move the arrow
  ; cursor out of the way if it's not already on the bottom row.
  lda cursor_y
  cmp #FIELD_HT-1
  bcs :+
  inc cursor_y
:
  ldx selectedCards
  bmi isNotMatch
  ldy selectedCards+1
  bmi isNotMatch
  lda boardState,x
  eor boardState,y
  and #$3F
  bne isNotMatch

  lda #4
  ldx isCombo
  beq :+
  lda #5
:
  jsr start_sound
  lda #1
  sta isCombo
  lda #PlayState::COLLECTING
  bne dgafMatch
isNotMatch:
  lda #2
  jsr start_sound
  lda #0
  sta isCombo
  lda #PlayState::UNFLIPPING

dgafMatch:
  sta curState
  lda #90
  sta stateTimer
  jmp nothingSelected
.endproc

.proc handleStateUnflipping
  lda stateTimer
  beq timeToUnflip
  dec stateTimer
  cmp #89
  bne notLoseScore
  lda scoreMethod
  bne notLoseScore
  ldx curTurn
  lda score,x
  beq notLoseScore
  dec score,x
  bne notGameOverSound
  
  ; play game over sound
  lda #$06
  sta bgcolor
  lda #7
  jsr start_sound
  lda #8
  jsr start_sound
  lda #PlayState::GAME_OVER
  sta curState
  ldx curTurn
notGameOverSound:
  ldy #0
  jsr buildScoreUpdate
notLoseScore:
  lda stateTimer
  cmp #6
  bcs noAnimateYet

  ; Start to unflip the cards
  cmp #5
  bne notFirstUnflipFrame
  lda #3
  jsr start_sound
  ldy selectedCards
  sty cardToDraw
  lda boardState,y
  and #$3F
  ora #$40
  sta boardState,y
  lda stateTimer
notFirstUnflipFrame:
  lsr a
  clc
  adc #$08
  sta card0FlipFrame
  rts
timeToUnflip:
  ldy selectedCards
  sty cardToDraw
  lda boardState,y
  and #$3F
  ora #$80
  sta boardState,y
  lda #$FF
  sta selectedCards
  lda selectedCards+1
  bpl secondCardNeedsUnflipped

  lda numPlayers
  cmp #2
  bcs multiplayer
  lda #0
  sta curAIState
  lda scoreMethod
  ora score
  bne notFailed
  lda #PlayState::DONE
  bne setCurState
notFailed:
  lda #PlayState::STILL
setCurState:
  sta curState
noAnimateYet:
  lda #0
  sta card0FlipFrame
  rts

secondCardNeedsUnflipped:
  sta selectedCards
  lda #$FF
  sta selectedCards+1
  lda #6
  sta stateTimer
  bne noAnimateYet

multiplayer:
  lda curTurn
  tax
  inx
  cpx numPlayers
  bcc noWrapAroundToFirstPlayer
  ldx #0
noWrapAroundToFirstPlayer: 
  stx curTurn
  ; hide old arrow
  tax
  ldy #0
  jsr buildScoreUpdate
  lda #PlayState::PASS_CONTROLLER
  sta curState
  bne noAnimateYet
.endproc

.proc handleStateCollecting
  lda stateTimer
  beq timeToUnflip
  dec stateTimer
  cmp #28
  beq start1Anim
  bcs noCoalesceYet
  cmp #16
  beq start2Anim
  bcs continue1Anim
  jmp clockCollecting2Animation
start1Anim:
  jsr initCollecting1Animation
  ldy selectedCards+1
  sty cardToDraw
  lda #$3F
  and boardState,y
  sta boardState,y
already1Anim:
  jsr clockCollecting1Animation
noCoalesceYet:
  lda #0
  sta card0FlipFrame
  rts
continue1Anim:
  ldy selectedCards
  sty cardToDraw
  lda #$3F
  and boardState,y
  sta boardState,y
  bpl already1Anim
start2Anim:
  jmp initCollecting2Animation
timeToUnflip:
  ldy selectedCards
  sty cardToDraw
  lda #0
  sta boardState,y
  lda #$FF
  sta selectedCards
  lda selectedCards+1
  bpl secondCardNeedsUnflipped

  ldx curTurn
  lda scoreMethod
  beq notAddScore
  inc score,x
notAddScore:
  ldy #0
  jsr buildScoreUpdate
  jsr clearCollectingAnimation
  jsr countRemainingCards
  cpy #0
  beq noCardsLeft
  lda numPlayers
  cmp #2
  bcc notVs
  
  ; Compute maximum possible score for other player
  ; as (nCardsLeft / 2) + otherPlayerCards
  lda curTurn
  eor #1
  tax
  tya
  lsr a
  clc
  adc score,x
  sta 0
  
  ; if score >= maxPossibleScore then give summary judgment.
  ldx curTurn
  lda score,x
  sec
  sbc 0
  bcs noCardsLeft

notVs:
  lda #0
  sta curAIState
  lda #PlayState::STILL
  sta curState
  rts
noCardsLeft:
  lda lastPlayerIsAI  ; no CPU player in the game: win for player.
  beq setWin
  ldx numPlayers  ; no human player in the game: win for CPU.
  dex
  beq setWin
  cpx curTurn  ; current player is not the CPU: win.
  bne setWin

  ; so if we have a cpu player and a human player, and it's the CPU
  ; player's turn, it's a lose.
  lda #90
  sta stateTimer
  lda #PlayState::GAME_OVER
  sta curState
  lda #$06
  sta bgcolor
  lda #7
  jsr start_sound
  lda #8
  jmp start_sound
setWin:
  lda #1
  jsr init_music
  lda #180
  sta stateTimer
  lda #PlayState::CLEARED
  sta curState
  rts
secondCardNeedsUnflipped:
  sta selectedCards
  lda #$FF
  sta selectedCards+1
  rts
.endproc

.proc handleStatePassController
  ldx curTurn
  ldy #0
  jsr buildScoreUpdate

  lda lastPlayerIsAI
  beq notPassingToAI
.ifdef WITH_FOUR_SCORE  ; deluxe version might allow >2 players
  lda nPlayers
  cmp #3
  bcc notPassingToAI
  clc
  sbc curTurn
  bne notPassingToAI
.endif
  lda #PlayState::STILL
  sta curState
  jmp done
notPassingToAI:

  ; find which keys
  ldx #1
padLoop:
  lda cur_keys,x
  and #KEY_A|KEY_B
  cmp #KEY_A|KEY_B
  bne notAPlusB
  and new_keys,x
  bne pressedAPlusB
notAPlusB:
  dex
  bpl padLoop
  rts
pressedAPlusB:
  stx activePad
  lda #1
  jsr start_sound

done:
  ; Cancel autorepeat so that keypresses from logging in don't
  ; leak into turning cards over. 
  ldx activePad
  lda #0
  sta das_keys,x
  lda #PlayState::STILL
  sta curState
  rts
.endproc

.proc handleStateGameOver
  lda stateTimer
  bne notDone
  lda #PlayState::DONE
  sta curState
notDone:
  dec stateTimer
  cmp #26
  bcc clearToCenter
  rts
clearToCenter:

  ; The playfield occupies rows 3 to 27 of the first nametable.
  ; We want to clear one row per frame: 
  lsr a      ; 12+ 12- 11+ 11- [...] 1+ 1- 0+
  bcc :+
  eor #$FF
:            ; 243+ 12- 244+ 11- [...] 254+ 1- 255+
  adc #15    ; 3 12 4 11 [...] 14 16 15
  sta gameOverClearTransitionY
  rts
.endproc

;;
; Counts the cards remaining in the field.
; @return y: number of cards
.proc countRemainingCards
  ldx #FIELD_WID*FIELD_HT-1
  ldy #0
lookForCardsLeft:
  lda boardState,x
  bpl :+
  iny
:
  dex
  bpl lookForCardsLeft
  rts
.endproc

